import json
import logging
import os
from typing import Optional

from pynumaflow.function import Messages, Datum

from numaprom.tools import catch_exception, msg_forward

_LOGGER = logging.getLogger(__name__)


@catch_exception
@msg_forward
def metric_filter(_: str, datum: Datum) -> Optional[Messages]:
    """
    UDF to filter metrics by labels
    """
    _LOGGER.debug("Received Msg: %s ", datum.value)

    msg = datum.value.decode("utf-8")
    data = json.loads(msg)

    label = os.getenv("LABEL")
    label_values = json.loads(os.getenv("LABEL_VALUES", "[]"))

    if label in data["labels"] and data["labels"][label] not in label_values:
        return None

    _LOGGER.info("Sending Metric: %s ", data)
    return msg



import logging
import os
import time
from collections import OrderedDict

import orjson
from numalogic.registry import MLflowRegistry
from pynumaflow.function import Datum

from numaprom._constants import DEFAULT_TRACKING_URI
from numaprom.entities import Status, StreamPayload, TrainerPayload
from numaprom.tools import msg_forward

_LOGGER = logging.getLogger(__name__)


def _load_artifact(payload: StreamPayload):
    registry = MLflowRegistry(
        tracking_uri=os.getenv("TRACKING_URI", DEFAULT_TRACKING_URI), artifact_type="sklearn"
    )
    return registry.load(
        skeys=[payload.composite_keys["namespace"], payload.composite_keys["name"]],
        dkeys=["preproc"],
    )


@msg_forward
def preprocess(_: str, datum: Datum) -> bytes:
    _start_time = time.perf_counter()
    _in_msg = datum.value.decode("utf-8")
    payload = StreamPayload(**orjson.loads(_in_msg))
    x_raw = payload.get_streamarray()

    _LOGGER.debug("%s - Received Payload: %r ", payload.uuid, payload)

    # Load artifact
    preproc_artifact = _load_artifact(payload)
    if not preproc_artifact:
        _LOGGER.info("%s - Preproc clf not found for %s", payload.uuid, payload.composite_keys)
        train_payload = TrainerPayload(
            uuid=payload.uuid, composite_keys=OrderedDict(payload.composite_keys)
        )
        return orjson.dumps(train_payload)

    # Perform preprocessing
    preproc_clf = preproc_artifact.artifact
    x_scaled = preproc_clf.transform(x_raw)

    # Prepare payload for forwarding
    payload.set_win_arr(x_scaled)
    payload.set_status(Status.PRE_PROCESSED)

    _LOGGER.info("%s - Sending Payload: %r ", payload.uuid, payload)
    _LOGGER.debug(
        "%s - Time taken in preprocess: %.4f sec", payload.uuid, time.perf_counter() - _start_time
    )
    return orjson.dumps(payload, option=orjson.OPT_SERIALIZE_NUMPY)



import json
import logging
from redis.cluster import RedisCluster

from numaprom.tools import is_host_reachable

_LOGGER = logging.getLogger(__name__)
redis_client: RedisCluster = None


def get_redis_client(
    host: str, port: str, password: str = None, decode_responses: bool = True, recreate=False
) -> RedisCluster:
    global redis_client

    if not recreate and redis_client:
        return redis_client

    redis_params = {
        "host": host,
        "port": port,
        "password": password,
        "decode_responses": decode_responses,
        "skip_full_coverage_check": True,
        "dynamic_startup_nodes": False,
    }
    _LOGGER.info("Redis params: %s", json.dumps(redis_params, indent=4))

    if not is_host_reachable(host, port):
        _LOGGER.error("Redis Cluster is unreachable after retries!")

    redis_client = RedisCluster(**redis_params)
    return redis_client



import logging
import time
from collections import OrderedDict
from datetime import datetime, timedelta
from typing import Dict, Optional

from numalogic.models.autoencoder import AutoencoderTrainer
from numalogic.registry import ArtifactData
from numalogic.tools.data import StreamingDataset
from orjson import orjson
from pynumaflow.function import Datum
from torch.utils.data import DataLoader

from numaprom.entities import Status, StreamPayload, TrainerPayload
from numaprom.tools import (
    load_model,
    get_metric_config,
    msgs_forward,
)

_LOGGER = logging.getLogger(__name__)


def _run_inference(
    payload: StreamPayload, artifact_data: ArtifactData, model_config: Dict
) -> StreamPayload:
    model = artifact_data.artifact
    stream_data = payload.get_streamarray()
    stream_loader = DataLoader(StreamingDataset(stream_data, model_config["win_size"]))

    trainer = AutoencoderTrainer()
    recon_err = trainer.predict(model, dataloaders=stream_loader)

    _LOGGER.info("%s - Successfully inferred", payload.uuid)

    payload.set_win_arr(recon_err.numpy())
    payload.set_status(Status.INFERRED)
    payload.set_metadata("version", artifact_data.extras.get("version"))
    return payload


def _get_model(payload: StreamPayload, model_config: dict) -> Optional[ArtifactData]:
    artifact_data = load_model(
        skeys=[payload.composite_keys["namespace"], payload.composite_keys["name"]],
        dkeys=[model_config["model_name"]],
    )
    if not artifact_data:
        payload.set_status(Status.ARTIFACT_NOT_FOUND)
        _LOGGER.info(
            "%s - Model not found for %s",
            payload.uuid,
            payload.composite_keys,
        )
        return None
    _LOGGER.debug("%s - Successfully loaded model from mlflow", payload.uuid)
    return artifact_data


def _is_model_stale(
    payload: StreamPayload, artifact_data: ArtifactData, model_config: dict
) -> bool:
    date_updated = artifact_data.extras["last_updated_timestamp"] / 1000
    stale_date = (
        datetime.now() - timedelta(hours=int(model_config["retrain_freq_hr"]))
    ).timestamp()
    if date_updated < stale_date:
        _LOGGER.info(
            "%s - Model found is stale for %s",
            payload.uuid,
            payload.composite_keys,
        )
        return True
    return False


@msgs_forward
def inference(_: str, datum: Datum) -> list[bytes]:
    _start_time = time.perf_counter()

    stream_payload = None
    train_payload = None
    _in_msg = datum.value.decode("utf-8")

    # Construct payload object
    try:
        stream_payload = StreamPayload(**orjson.loads(_in_msg))
    except TypeError:
        train_payload = TrainerPayload(**orjson.loads(_in_msg))

    # Check if trainer payload is passed on from previous vtx
    if train_payload:
        _LOGGER.debug("%s - Relaying forward trainer payload")
        return [orjson.dumps(train_payload)]

    messages = []

    # Load config
    metric_config = get_metric_config(stream_payload.composite_keys["name"])
    model_config = metric_config["model_config"]

    # Check if model exists
    artifact_data = _get_model(stream_payload, model_config)
    if not artifact_data:
        train_payload = TrainerPayload(
            uuid=stream_payload.uuid, composite_keys=OrderedDict(stream_payload.composite_keys)
        )
        return [orjson.dumps(train_payload)]

    # Check if current model is stale
    if _is_model_stale(stream_payload, artifact_data, model_config):
        train_payload = TrainerPayload(
            uuid=stream_payload.uuid, composite_keys=OrderedDict(stream_payload.composite_keys)
        )
        messages.append(orjson.dumps(train_payload))

    # Generate predictions
    payload = _run_inference(stream_payload, artifact_data, model_config)
    messages.append(orjson.dumps(payload, option=orjson.OPT_SERIALIZE_NUMPY))

    _LOGGER.info("%s - Sending Payload: %s ", payload.uuid, payload)
    _LOGGER.debug(
        "%s - Time taken in inference: %.4f sec", payload.uuid, time.perf_counter() - _start_time
    )
    return messages



import logging
import os
import time
from typing import List

import numpy as np
import pandas as pd
from numalogic.models.autoencoder import AutoencoderTrainer
from numalogic.models.autoencoder.variants import SparseVanillaAE
from numalogic.models.threshold import StdDevThreshold
from numalogic.tools.data import StreamingDataset
from orjson import orjson
from pynumaflow.sink import Datum, Responses, Response
from sklearn.preprocessing import StandardScaler
from torch.utils.data import DataLoader

from numaprom.entities import TrainerPayload
from numaprom.redis import get_redis_client
from numaprom.tools import get_metric_config, save_model, fetch_data

_LOGGER = logging.getLogger(__name__)

HOST = os.getenv("REDIS_HOST")
PORT = os.getenv("REDIS_PORT")
AUTH = os.getenv("REDIS_AUTH")
EXPIRY = int(os.getenv("REDIS_EXPIRY", 360))


# TODO: extract all good hashes, including when there are 2 hashes at a time
def clean_data(uuid: str, df: pd.DataFrame, hash_col: str, limit=12) -> pd.DataFrame:
    df.replace([np.inf, -np.inf], np.nan, inplace=True)
    df = df.fillna(method="ffill", limit=limit)
    df = df.fillna(method="bfill", limit=limit)
    if df.columns[df.isna().any()].tolist():
        df.dropna(inplace=True)

    if df.empty:
        return pd.DataFrame()
    df = df.reset_index()
    df = (
        pd.merge(df, df[df.duplicated("timestamp", keep=False)], indicator=True, how="outer")
        .query('_merge=="left_only"')
        .drop("_merge", axis=1)
    )
    df.set_index("timestamp", inplace=True)
    df.drop(hash_col, axis=1, inplace=True)
    df = df.sort_values(by=["timestamp"], ascending=True)
    if len(df) < (1.5 * 60 * 12):
        _LOGGER.exception("%s - Not enough training points to initiate training", uuid)
        return pd.DataFrame()
    return df


def _train_model(uuid, x, model_config):
    _start_train = time.perf_counter()

    win_size = model_config["win_size"]
    dataset = StreamingDataset(x, win_size)
    model = SparseVanillaAE(seq_len=win_size)

    trainer = AutoencoderTrainer(max_epochs=40)
    trainer.fit(model, train_dataloaders=DataLoader(dataset, batch_size=64))

    _LOGGER.debug(
        "%s - Time taken to train model: %.3f sec", uuid, time.perf_counter() - _start_train
    )

    train_reconerr = trainer.predict(model, dataloaders=DataLoader(dataset, batch_size=64))
    return train_reconerr.numpy(), model


def _preprocess(x_raw):
    clf = StandardScaler()
    x_scaled = clf.fit_transform(x_raw)
    return x_scaled, clf


def _find_threshold(x_reconerr):
    clf = StdDevThreshold()
    clf.fit(x_reconerr)
    return clf


def _is_new_request(payload: TrainerPayload) -> bool:
    redis_client = get_redis_client(HOST, PORT, password=AUTH, recreate=False)
    _ckeys = ":".join([payload.composite_keys["namespace"], payload.composite_keys["name"]])
    r_key = f"trainrollout::{_ckeys}"
    value = redis_client.get(r_key)
    if value:
        return False

    redis_client.setex(r_key, time=EXPIRY, value=1)
    return True


def train_rollout(datums: List[Datum]) -> Responses:
    responses = Responses()

    for _datum in datums:
        payload = TrainerPayload(**orjson.loads(_datum.value))

        _LOGGER.debug("%s - Starting Training for keys: %s", payload.uuid, payload.composite_keys)

        is_new = _is_new_request(payload)
        if not is_new:
            _LOGGER.debug(
                "%s - Skipping rollouts train request with keys: %s",
                payload.uuid,
                payload.composite_keys,
            )
            responses.append(Response.as_success(_datum.id))
            continue

        metric_config = get_metric_config(payload.composite_keys["name"])
        model_config = metric_config["model_config"]
        win_size = model_config["win_size"]

        train_df = fetch_data(
            payload,
            metric_config,
            {"namespace": payload.composite_keys["namespace"]},
            return_labels=["hash_id"],
        )
        try:
            train_df = clean_data(payload.uuid, train_df, "hash_id")
        except KeyError:
            _LOGGER.exception(
                "%s - KeyError while data cleaning for train payload: %s", payload.uuid, payload
            )
            responses.append(Response.as_success(_datum.id))
            continue

        if len(train_df) < model_config["win_size"]:
            _LOGGER.info(
                "%s - Skipping training since traindata size: %s is less than winsize: %s",
                payload.uuid,
                train_df.shape,
                win_size,
            )
            responses.append(Response.as_success(_datum.id))
            continue

        x_train, preproc_clf = _preprocess(train_df.to_numpy())
        x_reconerr, model = _train_model(payload.uuid, x_train, model_config)
        thresh_clf = _find_threshold(x_reconerr)

        # TODO change this to just use **composite_keys
        skeys = [payload.composite_keys["namespace"], payload.composite_keys["name"]]

        # TODO catch mlflow exception
        version = save_model(
            skeys=skeys, dkeys=["preproc"], model=preproc_clf, artifact_type="sklearn"
        )
        _LOGGER.info(
            "%s - Preproc model saved with skeys: %s with version: %s", payload.uuid, skeys, version
        )

        version = save_model(skeys=skeys, dkeys=[model_config["model_name"]], model=model)
        _LOGGER.info(
            "%s - Model saved with skeys: %s with version: %s", payload.uuid, skeys, version
        )

        version = save_model(
            skeys=skeys, dkeys=["thresh"], model=thresh_clf, artifact_type="sklearn"
        )
        _LOGGER.info(
            "%s - Threshold model saved with skeys: %s with version: %s",
            payload.uuid,
            skeys,
            version,
        )

        responses.append(Response.as_success(_datum.id))

    return responses



import logging
import requests
import numpy as np
import pandas as pd
from typing import Dict, List, Optional

_LOGGER = logging.getLogger(__name__)


class Prometheus:
    def __init__(self, prometheus_server: str):
        self.PROMETHEUS_SERVER = prometheus_server

    def query_metric(
        self,
        metric_name: str,
        start: float,
        end: float,
        labels_map: Dict = None,
        return_labels: List[str] = None,
        step: int = 30,
    ) -> pd.DataFrame:
        query = metric_name
        if labels_map:
            label_list = [str(key + "=" + "'" + labels_map[key] + "'") for key in labels_map]
            query = metric_name + "{" + ",".join(label_list) + "}"

        _LOGGER.debug("Prometheus Query: %s", query)

        if end < start:
            raise ValueError("end_time must not be before start_time")

        result = self.query_range(query, start, end, step)

        arr = np.array(result["values"])
        df = pd.DataFrame(arr, columns=["timestamp", metric_name])
        df = df.astype(float)

        data = result["metric"]
        if return_labels:
            for label in return_labels:
                if label in data:
                    df[label] = data[label]

        if not df.empty:
            df.set_index("timestamp", inplace=True)
            df.index = pd.to_datetime(df.index.astype(int), unit="s")
        return df

    def query_range(self, query: str, start: float, end: float, step: int = 30) -> Optional[Dict]:
        results = {}
        data_points = (end - start) / step
        temp_start = start
        while data_points > 11000:
            temp_end = temp_start + 11000 * step
            response = self.query_range_limit(query, temp_start, temp_end, step)
            if results:
                results["values"] = results["values"] + response["values"]
            else:
                results = response
            temp_start = temp_end
            data_points = (end - temp_start) / step

        if data_points > 0:
            response = self.query_range_limit(query, temp_start, end)
            if results:
                results["values"] = results["values"] + response["values"]
            else:
                _LOGGER.debug("Prometheus query has returned empty results.")
                results = response

        return results

    def query_range_limit(
        self, query: str, start: float, end: float, step: int = 30
    ) -> Optional[Dict]:
        data_points = (end - start) / step

        if data_points > 11000:
            _LOGGER.info("Limit query only supports 11,000 data points")
            return None

        results = None
        try:
            response = requests.get(
                self.PROMETHEUS_SERVER + "/api/v1/query_range",
                params={"query": query, "start": start, "end": end, "step": f"{step}s"},
            )
            results = response.json()["data"]["result"][0]
        except Exception as ex:
            _LOGGER.exception("Prometheus error: %r", ex)
        return results

    def query(self, query: str) -> Optional[Dict]:
        results = []
        try:
            response = requests.get(
                self.PROMETHEUS_SERVER + "/api/v1/query", params={"query": query}
            )
            if response:
                results = response.json()["data"]["result"]
            else:
                _LOGGER.debug("Prometheus query has returned empty results.")
        except Exception as ex:
            _LOGGER.exception("error: %r", ex)

        return results



import logging
import os
import time
from collections import OrderedDict

from numalogic.registry import MLflowRegistry
from orjson import orjson
from pynumaflow.function import Datum

from numaprom._constants import DEFAULT_TRACKING_URI, TRAIN_VTX_KEY, POSTPROC_VTX_KEY
from numaprom.entities import Status, StreamPayload, TrainerPayload
from numaprom.tools import conditional_forward

_LOGGER = logging.getLogger(__name__)


def _load_artifact(payload: StreamPayload):
    registry = MLflowRegistry(
        tracking_uri=os.getenv("TRACKING_URI", DEFAULT_TRACKING_URI), artifact_type="sklearn"
    )
    return registry.load(
        skeys=[payload.composite_keys["namespace"], payload.composite_keys["name"]],
        dkeys=["thresh"],
    )


@conditional_forward
def threshold(_: str, datum: Datum) -> list[tuple[str, bytes]]:
    _start_time = time.perf_counter()
    _in_msg = datum.value.decode("utf-8")

    stream_payload = None
    train_payload = None

    # Construct payload object
    try:
        stream_payload = StreamPayload(**orjson.loads(_in_msg))
    except TypeError:
        train_payload = TrainerPayload(**orjson.loads(_in_msg))

    # Check if trainer payload is passed on from previous vtx
    if train_payload:
        _LOGGER.debug("%s - Previous clf not found. Sending to trainer..", train_payload.uuid)
        return [(TRAIN_VTX_KEY, orjson.dumps(train_payload))]

    recon_err = stream_payload.get_streamarray()

    # Check if model exists
    thresh_artifact = _load_artifact(stream_payload)
    if not thresh_artifact:
        _LOGGER.info(
            "%s - Thresh clf not found for %s. Sending to trainer..",
            stream_payload.uuid,
            stream_payload.composite_keys,
        )
        train_payload = TrainerPayload(
            uuid=stream_payload.uuid, composite_keys=OrderedDict(stream_payload.composite_keys)
        )
        return [(TRAIN_VTX_KEY, orjson.dumps(train_payload))]

    # Calculate anomaly score
    thresh_clf = thresh_artifact.artifact
    y_score = thresh_clf.predict(recon_err)

    # Prepare payload for forwarding
    stream_payload.set_win_arr(y_score)
    stream_payload.set_status(Status.THRESHOLD)

    _LOGGER.info("%s - Sending Payload: %r ", stream_payload.uuid, stream_payload)
    _LOGGER.debug(
        "%s - Time taken in threshold: %.4f", stream_payload.uuid, time.perf_counter() - _start_time
    )
    return [(POSTPROC_VTX_KEY, orjson.dumps(stream_payload, option=orjson.OPT_SERIALIZE_NUMPY))]



import logging
import os
import time
from typing import List, Dict

import numpy as np
from numalogic.postprocess import TanhNorm
from orjson import orjson
from pynumaflow.function import Datum
from redis.exceptions import ConnectionError as RedisConnectionError

from numaprom.entities import Status, PrometheusPayload, StreamPayload
from numaprom.redis import get_redis_client
from numaprom.tools import msgs_forward, get_metric_config

_LOGGER = logging.getLogger(__name__)

HOST = os.getenv("REDIS_HOST")
PORT = os.getenv("REDIS_PORT")
AUTH = os.getenv("REDIS_AUTH")


def save_to_redis(payload: StreamPayload, final_score: float, recreate: bool):
    r = get_redis_client(HOST, PORT, password=AUTH, recreate=recreate)

    metric_name = payload.composite_keys["name"]
    metric_config = get_metric_config(metric_name)
    output_config = metric_config["output_config"]

    r_keys = payload.composite_keys
    r_keys.pop("name")

    metrics_list = output_config["unified_metrics"]
    r_key = f"{':'.join(r_keys.values())}:{payload.end_ts}"

    final_score = -1 if np.isnan(final_score) else final_score
    r.hset(r_key, mapping={metric_name: final_score})
    _LOGGER.debug(
        "%s - Saved to redis, redis_key: %s, metric: %s, anomaly_score: %d",
        payload.uuid,
        r_key,
        metric_name,
        final_score,
    )

    for m in metrics_list:
        if not r.hexists(name=r_key, key=m):
            _LOGGER.debug(
                "%s - Unable to generate unified anomaly, missing metric: %s, redis_key: %s",
                payload.uuid,
                m,
                r_key,
            )
            return -1, []

    _LOGGER.debug("%s - Received all metrics, generating unified anomaly", payload.uuid)
    max_anomaly = -1
    anomalies = []
    for m in metrics_list:
        val = float(r.hget(name=r_key, key=m))
        anomalies.append(val)
        if max_anomaly < val:
            max_anomaly = val
    r.delete(r_key)

    return max_anomaly, anomalies


def __construct_publisher_payload(
    stream_payload: StreamPayload, final_score: float
) -> PrometheusPayload:
    metric_name = stream_payload.composite_keys["name"]
    namespace = stream_payload.composite_keys["namespace"]

    labels = {
        "model_version": str(stream_payload.get_metadata("version")),
        "namespace": stream_payload.get_metadata("src_labels").get("namespace"),
    }
    subsystem = stream_payload.get_metadata("src_labels").get("hash_id") or None

    return PrometheusPayload(
        timestamp_ms=int(stream_payload.end_ts),
        name=f"{metric_name}_anomaly",
        namespace=namespace,
        subsystem=subsystem,
        type="Gauge",
        value=float(final_score),
        labels=labels,
    )


def __construct_unified_payload(
    stream_payload: StreamPayload, max_anomaly: float, output_config: Dict
) -> PrometheusPayload:
    namespace = stream_payload.composite_keys["namespace"]

    labels = {
        "model_version": str(stream_payload.get_metadata("version")),
        "namespace": stream_payload.get_metadata("src_labels").get("namespace"),
        "hash_id": stream_payload.get_metadata("src_labels").get("hash_id"),
    }

    subsystem = stream_payload.get_metadata("src_labels").get("hash_id") or None

    return PrometheusPayload(
        timestamp_ms=int(stream_payload.end_ts),
        name=output_config["unified_metric_name"],
        namespace=namespace,
        subsystem=subsystem,
        type="Gauge",
        value=max_anomaly,
        labels=labels,
    )


def _publish(final_score: float, payload: StreamPayload) -> List[bytes]:
    metric_name = payload.composite_keys["name"]
    model_config = get_metric_config(metric_name)["model_config"]
    output_config = get_metric_config(metric_name)["output_config"]

    publisher_json = __construct_publisher_payload(payload, final_score).as_json()
    _LOGGER.info("%s - Payload sent to publisher: %s", payload.uuid, publisher_json)

    if model_config["name"] == "default":
        _LOGGER.debug(
            "%s - Using default config, cannot generate a unified anomaly score", payload.uuid
        )
        return [publisher_json]

    try:
        max_anomaly, anomalies = save_to_redis(
            payload=payload, final_score=final_score, recreate=False
        )
    except RedisConnectionError:
        _LOGGER.warning("%s - Redis connection failed, recreating the redis client", payload.uuid)
        max_anomaly, anomalies = save_to_redis(
            payload=payload, final_score=final_score, recreate=True
        )

    if max_anomaly > -1:
        unified_json = __construct_unified_payload(payload, max_anomaly, output_config).as_json()
        _LOGGER.info(
            "%s - Unified anomaly payload sent to publisher: %s", payload.uuid, unified_json
        )
        return [publisher_json, unified_json]
    return [publisher_json]


@msgs_forward
def postprocess(_: str, datum: Datum) -> List[bytes]:
    _start_time = time.perf_counter()

    _in_msg = datum.value.decode("utf-8")
    payload = StreamPayload(**orjson.loads(_in_msg))

    _LOGGER.debug("%s - Received Payload: %r ", payload.uuid, payload)

    raw_scores = payload.get_streamarray()
    raw_mean_score = np.mean(raw_scores)

    postproc_clf = TanhNorm()
    norm_score = postproc_clf.transform(raw_mean_score)

    payload.set_status(Status.POST_PROCESSED)
    _LOGGER.info("%s - Successfully post-processed; final score: %s", payload.uuid, norm_score)

    _LOGGER.debug(
        "%s - Time taken in postprocess: %.4f sec", payload.uuid, time.perf_counter() - _start_time
    )
    return _publish(norm_score, payload)



import logging
import os
import time
import uuid
from typing import List, Tuple, Optional

import numpy as np
import numpy.typing as npt
from orjson import orjson
from pynumaflow.function import Datum
from redis.exceptions import ConnectionError as RedisConnectionError

from numaprom.entities import StreamPayload, Status
from numaprom.redis import get_redis_client
from numaprom.tools import msg_forward, create_composite_keys, get_metric_config

_LOGGER = logging.getLogger(__name__)

HOST = os.getenv("REDIS_HOST")
PORT = os.getenv("REDIS_PORT")
AUTH = os.getenv("REDIS_AUTH")


# TODO get the replace value from config
def _clean_arr(
    id_: str,
    ckeys: dict,
    arr: npt.NDArray[float],
    replace_val: float = 0.0,
    inf_replace: float = 1e10,
) -> npt.NDArray[float]:
    if not np.isfinite(arr).any():
        _LOGGER.warning(
            "%s - Non finite values encountered: %s for keys: %s", id_, list(arr), ckeys
        )
    return np.nan_to_num(arr, nan=replace_val, posinf=inf_replace, neginf=-inf_replace)


def __aggregate_window(key, ts, value, win_size, buff_size, recreate) -> List[Tuple[float, float]]:
    redis_client = get_redis_client(HOST, PORT, password=AUTH, recreate=recreate)
    with redis_client.pipeline() as pl:
        pl.zadd(key, {f"{value}::{ts}": ts})
        pl.zremrangebyrank(key, -(buff_size + 10), -buff_size)
        pl.zrange(key, -win_size, -1, withscores=True, score_cast_func=int)
        out = pl.execute()
    _window = out[-1]
    _window = list(map(lambda x: (float(x[0].split("::")[0]), x[1]), _window))
    return _window


@msg_forward
def window(_: str, datum: Datum) -> Optional[bytes]:
    """
    UDF to construct windowing of the streaming input data, required by ML models.
    """
    _LOGGER.debug("Received Msg: %s ", datum.value)

    _start_time = time.perf_counter()
    msg = orjson.loads(datum.value)

    metric_name = msg["name"]
    metric_config = get_metric_config(metric_name)
    win_size = metric_config["model_config"]["win_size"]
    buff_size = int(os.getenv("BUFF_SIZE", 10 * win_size))

    if buff_size < win_size:
        raise ValueError(
            f"Redis list buffer size: {buff_size} is less than window length: {win_size}"
        )

    key_map = create_composite_keys(msg)
    unique_key = ":".join(key_map.values())
    value = float(msg["value"])

    try:
        elements = __aggregate_window(
            unique_key, msg["timestamp"], value, win_size, buff_size, recreate=False
        )
    except RedisConnectionError:
        _LOGGER.warning("Redis connection failed, recreating the redis client")
        elements = __aggregate_window(
            unique_key, msg["timestamp"], value, win_size, buff_size, recreate=True
        )

    if len(elements) < win_size:
        return None

    win_list = [float(_val) for _val, _ in elements]
    win_arr = np.asarray(win_list).reshape(-1, 1)
    _uuid = uuid.uuid4().hex
    composite_keys = create_composite_keys(msg)

    payload = StreamPayload(
        uuid=uuid.uuid4().hex,
        composite_keys=composite_keys,
        status=Status.EXTRACTED,
        win_arr=_clean_arr(_uuid, composite_keys, win_arr),
        win_ts_arr=[str(_ts) for _, _ts in elements],
        metadata=dict(src_labels=msg["labels"]),
    )

    _LOGGER.info("%s - Sending Payload: %r ", payload.uuid, payload)
    _LOGGER.debug(
        "%s - Time taken in window: %.4f sec", payload.uuid, time.perf_counter() - _start_time
    )
    return orjson.dumps(payload, option=orjson.OPT_SERIALIZE_NUMPY)



import logging
import os

LOGGER = logging.getLogger(__name__)

stream_handler = logging.StreamHandler()

if os.getenv("DEBUG", False):
    LOGGER.setLevel(logging.DEBUG)
    stream_handler.setLevel(logging.DEBUG)
else:
    LOGGER.setLevel(logging.INFO)
    stream_handler.setLevel(logging.INFO)

formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
stream_handler.setFormatter(formatter)


LOGGER.addHandler(stream_handler)
logging.getLogger("pytorch_lightning").setLevel(logging.WARNING)



import logging
import os
import time
from typing import List

import numpy as np
import pandas as pd
from numalogic.models.autoencoder import AutoencoderTrainer
from numalogic.models.autoencoder.variants import SparseVanillaAE
from numalogic.models.threshold import StdDevThreshold
from numalogic.tools.data import StreamingDataset
from orjson import orjson
from pynumaflow.sink import Datum, Responses, Response
from sklearn.preprocessing import StandardScaler
from torch.utils.data import DataLoader

from numaprom.entities import TrainerPayload
from numaprom.redis import get_redis_client
from numaprom.tools import get_metric_config, save_model, fetch_data

_LOGGER = logging.getLogger(__name__)

HOST = os.getenv("REDIS_HOST")
PORT = os.getenv("REDIS_PORT")
AUTH = os.getenv("REDIS_AUTH")
EXPIRY = int(os.getenv("REDIS_EXPIRY", 300))


def clean_data(df: pd.DataFrame, limit=12) -> pd.DataFrame:
    df.replace([np.inf, -np.inf], np.nan, inplace=True)
    df = df.fillna(method="ffill", limit=limit)
    df = df.fillna(method="bfill", limit=limit)
    if df.columns[df.isna().any()].tolist():
        df.dropna(inplace=True)
    return df


def _train_model(uuid, x, model_config):
    _start_train = time.time()

    win_size = model_config["win_size"]
    dataset = StreamingDataset(x, win_size)
    model = SparseVanillaAE(seq_len=win_size)

    trainer = AutoencoderTrainer(max_epochs=40)
    trainer.fit(model, train_dataloaders=DataLoader(dataset, batch_size=64))

    _LOGGER.debug("%s - Time taken to train model: %s", uuid, time.perf_counter() - _start_train)

    train_reconerr = trainer.predict(model, dataloaders=DataLoader(dataset, batch_size=64))
    return train_reconerr.numpy(), model


def _preprocess(x_raw):
    clf = StandardScaler()
    x_scaled = clf.fit_transform(x_raw)
    return x_scaled, clf


def _find_threshold(x_reconerr):
    clf = StdDevThreshold()
    clf.fit(x_reconerr)
    return clf


def _is_new_request(payload: TrainerPayload) -> bool:
    redis_client = get_redis_client(HOST, PORT, password=AUTH, recreate=False)
    _ckeys = ":".join([payload.composite_keys["namespace"], payload.composite_keys["name"]])
    r_key = f"train::{_ckeys}"

    value = redis_client.get(r_key)
    if value:
        return False

    redis_client.setex(r_key, time=EXPIRY, value=1)
    return True


def train(datums: List[Datum]) -> Responses:
    responses = Responses()

    for _datum in datums:
        payload = TrainerPayload(**orjson.loads(_datum.value))

        _LOGGER.debug("%s - Starting Training for keys: %s", payload.uuid, payload.composite_keys)

        is_new = _is_new_request(payload)
        if not is_new:
            _LOGGER.debug(
                "%s - Skipping train request with keys: %s", payload.uuid, payload.composite_keys
            )
            responses.append(Response.as_success(_datum.id))
            continue

        metric_config = get_metric_config(payload.composite_keys["name"])
        model_config = metric_config["model_config"]
        win_size = model_config["win_size"]

        train_df = fetch_data(
            payload, metric_config, {"namespace": payload.composite_keys["namespace"]}
        )
        train_df = clean_data(train_df)

        if len(train_df) < model_config["win_size"]:
            _LOGGER.info(
                "%s - Skipping training since traindata size: %s is less than winsize: %s",
                payload.uuid,
                train_df.shape,
                win_size,
            )
            responses.append(Response.as_success(_datum.id))
            continue

        x_train, preproc_clf = _preprocess(train_df.to_numpy())
        x_reconerr, model = _train_model(payload.uuid, x_train, model_config)
        thresh_clf = _find_threshold(x_reconerr)

        # TODO change this to just use **composite_keys
        skeys = [payload.composite_keys["namespace"], payload.composite_keys["name"]]

        version = save_model(
            skeys=skeys, dkeys=["preproc"], model=preproc_clf, artifact_type="sklearn"
        )
        _LOGGER.info(
            "%s - Preproc model saved with skeys: %s with version: %s", payload.uuid, skeys, version
        )

        version = save_model(skeys=skeys, dkeys=[model_config["model_name"]], model=model)
        _LOGGER.info(
            "%s - Model saved with skeys: %s with version: %s", payload.uuid, skeys, version
        )

        version = save_model(
            skeys=skeys, dkeys=["thresh"], model=thresh_clf, artifact_type="sklearn"
        )
        _LOGGER.info(
            "%s - Threshold model saved with skeys: %s with version: %s",
            payload.uuid,
            skeys,
            version,
        )

        responses.append(Response.as_success(_datum.id))

    return responses



import os

NUMAPROM_DIR = os.path.dirname(__file__)
ROOT_DIR = os.path.split(NUMAPROM_DIR)[0]
TESTS_DIR = os.path.join(ROOT_DIR, "tests")
DATA_DIR = os.path.join(NUMAPROM_DIR, "data")

# endpoints
DEFAULT_PROMETHEUS_SERVER = "http://prometheus-service.monitoring.svc.cluster.local:8080"
DEFAULT_TRACKING_URI = "http://mlflow-service.numalogic-prometheus.svc.cluster.local:5000"


# UDF constants
TRAIN_VTX_KEY = "train"
INFERENCE_VTX_KEY = "inference"
THRESHOLD_VTX_KEY = "threshold"
POSTPROC_VTX_KEY = "postproc"


OUTPUT_CONFIG = {
    "argo_cd": {
        "unified_strategy": "max",
        "unified_metric_name": "namespace_app_pod_http_server_requests_unified_anomaly",
        "unified_metrics": [
            "namespace_app_pod_http_server_requests_errors",
            "namespace_app_pod_http_server_requests_error_rate",
            "namespace_app_pod_http_server_requests_latency",
            "namespace_asset_pod_cpu_utilization",
            "namespace_asset_pod_memory_utilization",
        ],
    },
    "argo_rollouts": {
        "unified_strategy": "max",
        "unified_metric_name": "namespace_hash_pod_http_server_requests_unified_anomaly",
        "unified_metrics": [
            "namespace_hash_pod_http_server_requests_error_rate",
            "namespace_hash_pod_http_server_requests_latency",
        ],
    },
    "fuzzy_rollouts": {
        "unified_strategy": "max",
        "unified_metric_name": "namespace_rollout_api_unified_anomaly",
        "unified_metrics": [
            "namespace_rollout_api_error_rate",
        ],
    },
    "default": {"unified_strategy": None, "unified_metric_name": None, "unified_metrics": None},
}

# ML parameters
MODEL_CONFIG = {
    "argo_cd": {
        "name": "argo_cd",
        "win_size": 12,
        "threshold_min": 0.1,
        "model_name": "ae_sparse",
        "retrain_freq_hr": 8,
        "resume_training": "True",
        "num_epochs": 100,
        "training_keys": ["namespace", "name"],
    },
    "argo_rollouts": {
        "name": "argo_rollouts",
        "win_size": 12,
        "threshold_min": 0.001,
        "model_name": "ae_sparse",
        "retrain_freq_hr": 8,
        "resume_training": "True",
        "num_epochs": 50,
        "training_keys": ["namespace", "name"],
    },
    "fuzzy_rollouts": {
        "name": "fuzzy_rollouts",
        "win_size": 12,
        "threshold_min": 0.001,
        "model_name": "ae_sparse",
        "retrain_freq_hr": 8,
        "resume_training": "True",
        "num_epochs": 50,
        "training_keys": ["namespace", "name"],
    },
    "default": {
        "name": "default",
        "win_size": 12,
        "threshold_min": 0.001,
        "model_name": "ae_sparse",
        "retrain_freq_hr": 8,
        "resume_training": "True",
        "num_epochs": 50,
        "training_keys": ["namespace", "name"],
    },
}

METRIC_CONFIG = {
    "namespace_app_pod_http_server_requests_errors": {
        "keys": ["namespace", "name"],
        "scrape_interval": 5,
        "model_config": MODEL_CONFIG["argo_cd"],
        "output_config": OUTPUT_CONFIG["argo_cd"],
    },
    "namespace_app_pod_http_server_requests_error_rate": {
        "keys": ["namespace", "name"],
        "scrape_interval": 5,
        "model_config": MODEL_CONFIG["argo_cd"],
        "output_config": OUTPUT_CONFIG["argo_cd"],
    },
    "namespace_app_pod_http_server_requests_latency": {
        "keys": ["namespace", "name"],
        "scrape_interval": 5,
        "model_config": MODEL_CONFIG["argo_cd"],
        "output_config": OUTPUT_CONFIG["argo_cd"],
    },
    "namespace_asset_pod_cpu_utilization": {
        "keys": ["namespace", "name"],
        "scrape_interval": 5,
        "model_config": MODEL_CONFIG["argo_cd"],
        "output_config": OUTPUT_CONFIG["argo_cd"],
    },
    "namespace_asset_pod_memory_utilization": {
        "keys": ["namespace", "name"],
        "scrape_interval": 5,
        "model_config": MODEL_CONFIG["argo_cd"],
        "output_config": OUTPUT_CONFIG["argo_cd"],
    },
    "namespace_hash_pod_http_server_requests_error_rate": {
        "keys": ["namespace", "name", "hash_id"],
        "scrape_interval": 5,
        "model_config": MODEL_CONFIG["argo_rollouts"],
        "output_config": OUTPUT_CONFIG["argo_rollouts"],
    },
    "namespace_hash_pod_http_server_requests_latency": {
        "keys": ["namespace", "name", "hash_id"],
        "scrape_interval": 5,
        "model_config": MODEL_CONFIG["argo_rollouts"],
        "output_config": OUTPUT_CONFIG["argo_rollouts"],
    },
    "namespace_rollout_api_latency": {
        "keys": ["namespace", "name", "hash_id"],
        "scrape_interval": 30,
        "model_config": MODEL_CONFIG["fuzzy_rollouts"],
        "output_config": OUTPUT_CONFIG["fuzzy_rollouts"],
    },
    "namespace_rollout_api_error_rate": {
        "keys": ["namespace", "name", "hash_id"],
        "scrape_interval": 30,
        "model_config": MODEL_CONFIG["fuzzy_rollouts"],
        "output_config": OUTPUT_CONFIG["fuzzy_rollouts"],
    },
    "default": {
        "keys": ["namespace", "name", "hash_id"],
        "scrape_interval": 30,
        "model_config": MODEL_CONFIG["default"],
        "output_config": OUTPUT_CONFIG["default"],
    },
}



from copy import copy
from dataclasses import dataclass
from enum import Enum
from typing import List, Dict, Optional, Any, Union, OrderedDict

import numpy as np
import numpy.typing as npt
import orjson

Vector = List[float]
Matrix = Union[Vector, List[Vector], npt.NDArray[float]]


class Status(str, Enum):
    RAW = "raw"
    EXTRACTED = "extracted"
    PRE_PROCESSED = "pre_processed"
    INFERRED = "inferred"
    THRESHOLD = "threshold_complete"
    POST_PROCESSED = "post_processed"
    ARTIFACT_NOT_FOUND = "artifact_not_found"
    ARTIFACT_STALE = "artifact_is_stale"


@dataclass(frozen=True)
class TrainerPayload:
    uuid: str
    composite_keys: OrderedDict[str, str]


@dataclass(repr=False)
class StreamPayload:
    uuid: str
    win_arr: Matrix
    win_ts_arr: List[str]
    composite_keys: Dict[str, str]
    status: Status = Status.RAW
    metadata: Dict[str, Any] = None

    @property
    def start_ts(self):
        return self.win_ts_arr[0]

    @property
    def end_ts(self):
        return self.win_ts_arr[-1]

    def get_streamarray(self):
        return np.asarray(self.win_arr)

    def get_metadata(self, key: str):
        return copy(self.metadata[key])

    def set_win_arr(self, arr):
        self.win_arr = arr

    def set_status(self, status: Status):
        self.status = status

    def set_metadata(self, key: str, value):
        self.metadata[key] = value

    def __repr__(self):
        return "{uuid: %s, win_arr: %s, win_ts_arr: %s, composite_keys: %s, metadata: %s}" % (
            self.uuid,
            list(self.win_arr),
            self.win_ts_arr,
            self.composite_keys,
            self.metadata,
        )


@dataclass(repr=False)
class PrometheusPayload:
    timestamp_ms: int
    name: str
    namespace: str
    subsystem: Optional[str]
    type: str
    value: float
    labels: Dict[str, str]

    def as_json(self) -> bytes:
        return orjson.dumps(
            {
                "TimestampMs": self.timestamp_ms,
                "Name": self.name,
                "Namespace": self.namespace,
                "Subsystem": self.subsystem,
                "Type": self.type,
                "Value": self.value,
                "Labels": self.labels,
            }
        )

    @classmethod
    def from_json(cls, json_obj) -> "PrometheusPayload":
        obj = orjson.loads(json_obj)
        return cls(
            timestamp_ms=obj["TimestampMs"],
            name=obj["Name"],
            namespace=obj["Namespace"],
            subsystem=obj["Subsystem"],
            type=obj["Type"],
            value=obj["Value"],
            labels=obj["Labels"],
        )

    def __repr__(self):
        return (
            "{timestamp_ms: %s, name: %s, namespace: %s, "
            "subsystem: %s, type: %s, value: %s, labels: %s}"
            % (
                self.timestamp_ms,
                self.name,
                self.namespace,
                self.subsystem,
                self.type,
                self.value,
                self.labels,
            )
        )



from numaprom.udf.filter import metric_filter
from numaprom.udf.inference import inference
from numaprom.udf.postprocess import postprocess
from numaprom.udf.preprocess import preprocess
from numaprom.udf.window import window
from numaprom.udf.threshold import threshold


__all__ = ["preprocess", "metric_filter", "inference", "window", "postprocess", "threshold"]



from typing import Callable, Union

from pynumaflow.function import Messages
from pynumaflow.sink import Responses

from numaprom.udf import preprocess, postprocess, window, metric_filter, inference, threshold
from numaprom.udsink import train, train_rollout


class HandlerFactory:
    @classmethod
    def get_handler(cls, step: str) -> Callable[..., Union[Messages, Responses]]:
        if step == "metric_filter":
            return metric_filter

        if step == "window":
            return window

        if step == "preprocess":
            return preprocess

        if step == "inference":
            return inference

        if step == "postprocess":
            return postprocess

        if step == "threshold":
            return threshold

        if step == "train":
            return train

        if step == "train_rollout":
            return train_rollout

        raise NotImplementedError(f"Invalid step provided: {step}")



import logging
import os
import socket
import time
from datetime import timedelta, datetime
from functools import wraps
from json import JSONDecodeError
from typing import Optional, Dict, Sequence, Any

import pandas as pd
import pytz
from mlflow.entities.model_registry import ModelVersion
from mlflow.exceptions import RestException
from numalogic.registry import MLflowRegistry, ArtifactData
from pynumaflow.function import Messages, Message

from numaprom._constants import DEFAULT_TRACKING_URI, METRIC_CONFIG, DEFAULT_PROMETHEUS_SERVER
from numaprom.entities import TrainerPayload
from numaprom.prometheus import Prometheus

_LOGGER = logging.getLogger(__name__)


def catch_exception(func):
    @wraps(func)
    def inner_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except JSONDecodeError as err:
            _LOGGER.exception("Error in json decode for %s: %r", func.__name__, err)
        except Exception as ex:
            _LOGGER.exception("Error in %s: %r", func.__name__, ex)

    return inner_function


def msgs_forward(handler_func):
    @wraps(handler_func)
    def inner_function(*args, **kwargs):
        json_list = handler_func(*args, **kwargs)
        msgs = Messages()
        for json_data in json_list:
            if json_data:
                msgs.append(Message.to_all(json_data))
            else:
                msgs.append(Message.to_drop())
        return msgs

    return inner_function


def msg_forward(handler_func):
    @wraps(handler_func)
    def inner_function(*args, **kwargs):
        json_data = handler_func(*args, **kwargs)
        msgs = Messages()
        if json_data:
            msgs.append(Message.to_all(value=json_data))
        else:
            msgs.append(Message.to_drop())
        return msgs

    return inner_function


def conditional_forward(hand_func):
    @wraps(hand_func)
    def inner_function(*args, **kwargs) -> Messages:
        data = hand_func(*args, **kwargs)
        msgs = Messages()
        for vertex, json_data in data:
            if json_data and vertex:
                msgs.append(Message.to_vtx(key=vertex.encode(), value=json_data))
            else:
                msgs.append(Message.to_drop())
        return msgs

    return inner_function


def create_composite_keys(msg: dict) -> Dict:
    labels = msg.get("labels")
    metric_name = msg["name"]

    keys = get_metric_config(metric_name)["keys"]
    result = {}
    for k in keys:
        if k in msg:
            result[k] = msg[k]
        if k in labels:
            result[k] = labels[k]
    return result


def get_ipv4_by_hostname(hostname: str, port=0) -> list:
    return list(
        idx[4][0]
        for idx in socket.getaddrinfo(hostname, port)
        if idx[0] is socket.AddressFamily.AF_INET and idx[1] is socket.SocketKind.SOCK_RAW
    )


def is_host_reachable(hostname: str, port=None, max_retries=5, sleep_sec=5) -> bool:
    retries = 0
    assert max_retries >= 1, "Max retries has to be at least 1"

    while retries < max_retries:
        try:
            get_ipv4_by_hostname(hostname, port)
        except socket.gaierror as ex:
            retries += 1
            _LOGGER.warning(
                "Failed to resolve hostname: %s: error: %r", hostname, ex, exc_info=True
            )
            time.sleep(sleep_sec)
        else:
            return True
    _LOGGER.error("Failed to resolve hostname: %s even after retries!")
    return False


def load_model(skeys: Sequence[str], dkeys: Sequence[str]) -> Optional[ArtifactData]:
    try:
        tracking_uri = os.getenv("TRACKING_URI", DEFAULT_TRACKING_URI)
        ml_registry = MLflowRegistry(tracking_uri=tracking_uri)
        return ml_registry.load(skeys=skeys, dkeys=dkeys)
    except RestException as warn:
        if warn.error_code == 404:
            return None
        _LOGGER.warning("Non 404 error from mlflow: %r", warn)
    except Exception as ex:
        _LOGGER.error("Unexpected error while loading model from MLflow database: %r", ex)
        return None


def save_model(
    skeys: Sequence[str], dkeys: Sequence[str], model, artifact_type="pytorch", **metadata
) -> Optional[ModelVersion]:
    tracking_uri = os.getenv("TRACKING_URI", DEFAULT_TRACKING_URI)
    ml_registry = MLflowRegistry(tracking_uri=tracking_uri, artifact_type=artifact_type)
    version = ml_registry.save(skeys=skeys, dkeys=dkeys, artifact=model, **metadata)
    return version


def get_metric_config(metric_name: str) -> Dict[str, Any]:
    if metric_name in METRIC_CONFIG:
        return METRIC_CONFIG[metric_name]
    return METRIC_CONFIG["default"]


def fetch_data(
    payload: TrainerPayload, metric_config: dict, labels: dict, return_labels=None
) -> pd.DataFrame:
    _start_time = time.time()

    prometheus_server = os.getenv("PROMETHEUS_SERVER", DEFAULT_PROMETHEUS_SERVER)
    datafetcher = Prometheus(prometheus_server)

    end_dt = datetime.now(pytz.utc)
    start_dt = end_dt - timedelta(hours=15)

    df = datafetcher.query_metric(
        metric_name=payload.composite_keys["name"],
        labels_map=labels,
        return_labels=return_labels,
        start=start_dt.timestamp(),
        end=end_dt.timestamp(),
        step=metric_config["scrape_interval"],
    )
    _LOGGER.info(
        "%s - Time taken to fetch data: %s, for df shape: %s",
        payload.uuid,
        time.time() - _start_time,
        df.shape,
    )
    return df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['numaprom', 'numaprom.udf', 'numaprom.udsink']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.25.2,<2.0.0',
 'numalogic[mlflow]>=0.3.1,<0.4.0',
 'orjson>=3.8.4,<4.0.0',
 'pynumaflow>=0.2.6,<0.3.0',
 'redis>=4.3.1,<5.0.0']

setup_kwargs = {
    'name': 'numalogic-prometheus',
    'version': '0.1.5a4',
    'description': 'ML inference on numaflow using numalogic on Prometheus metrics',
    'long_description': 'None',
    'author': 'Numalogic developers',
    'author_email': 'None',
    'maintainer': 'Avik Basu',
    'maintainer_email': 'avikbasu93@gmail.com',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
